import pytest
import os
from pyway.checksum import Checksum
from pyway.migrate import Migrate
from pyway.settings import ConfigFile

from mysqld_integration_test import Mysqld


@pytest.fixture
def mysqld_connect(autouse: bool = True) -> Mysqld:
    mysqld = Mysqld()
    return mysqld.run()


@pytest.mark.checksum_test
@pytest.mark.mysqld_test
def test_pyway_table_checksum(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.checksum_file = "V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "8327AD7B"


@pytest.mark.checksum_test
@pytest.mark.mysqld_test
def test_pyway_table_checksum_fileinvalid(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(AttributeError):
        _, _ = Checksum(config).run()

    assert True


@pytest.mark.checksum_test
@pytest.mark.mysqld_test
def test_pyway_table_checksum_fullpath(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.checksum_file = "schema/V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "8327AD7B"


@pytest.mark.checksum_test
@pytest.mark.mysqld_test
def test_pyway_table_checksum_invalid_filename(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.checksum_file = "invalidfilename.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(FileNotFoundError):
        _, _ = Checksum(config).run()

    assert True



from pyway.helpers import Utils
from typing import List, Any, Optional, Type


class Migration():
    def __init__(self, version: Any, extension: Any, name: Any,
                 checksum: Any, apply_timestamp: Optional[Any]) -> None:
        self.version: str = version
        self.extension: str = extension
        self.name: str = name
        self.checksum: str = checksum
        self.apply_timestamp: Optional[Any] = apply_timestamp

    @classmethod
    def from_name(cls: Type['Migration'], name: str, path: str, **kwargs: str) -> 'Migration':
        version = kwargs.get('version', Utils.get_version_from_name(name))
        extension = kwargs.get('extension', Utils.get_extension_from_name(name))
        checksum = kwargs.get('checksum', Utils.load_checksum_from_name(name, path))
        apply_timestamp = kwargs.get('apply_timestamp')
        return cls(version, extension, name, checksum, apply_timestamp)

    @classmethod
    def from_list(cls, list_: List['Migration']) -> List['Migration']:
        return [cls(m.version, m.extension, m.name, m.checksum, m.apply_timestamp) for m in list_]

    def __str__(self) -> str:
        return f"version={self.version}, extension={self.extension}, name={self.name}, " \
               f"checksum={self.checksum}, apply_timestamp={self.apply_timestamp}"



import pytest
import os
from strip_ansi import strip_ansi
from pyway.info import Info
from pyway.settings import ConfigFile

from postgresql_integration_test import PostgreSQL

INFO_OUTPUT = """+-----------+-------------+-------------------+------------+-------------------+
|   version | extension   | name              | checksum   | apply_timestamp   |
|-----------+-------------+-------------------+------------+-------------------|
|      1.01 | SQL         | V01_01__test1.sql | new        | new               |
|      1.02 | SQL         | V01_02__test2.sql | new        | new               |
|      1.03 | SQL         | V01_03__test3.sql | new        | new               |
+-----------+-------------+-------------------+------------+-------------------+"""


@pytest.fixture
def postgresql_connect(autouse=True) -> PostgreSQL:
    postgresql = PostgreSQL()
    return postgresql.run()


@pytest.mark.info_test
@pytest.mark.postgresql_test
def test_pyway_info(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    tbl = Info(config).run()
    assert strip_ansi(tbl) == INFO_OUTPUT


@pytest.mark.info_test
@pytest.mark.postgresql_test
def test_pyway_info_nofiles(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    files = Info(config).get_new_local_migrations([], config.database_migration_dir)
    assert files == []



import pytest
import os
from pyway.helpers import Utils
from pyway.migration import Migration


@pytest.mark.helpers_test
def test_get_local_files() -> None:
    files = Utils.get_local_files(os.path.join('tests', 'data', 'schema'))
    assert len(files) == 3


@pytest.mark.helpers_test
def test_get_local_files_notfound() -> None:
    with pytest.raises(Exception):
        _ = Utils.get_local_files(os.path.join('tests', 'datanotfound'))
    assert True


@pytest.mark.helpers_test
def test_subtract_result() -> None:
    a = [Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema')),
         Migration.from_name('V01_02__test2.sql', os.path.join('tests', 'data', 'schema'))]
    b = [Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema'))]
    c = [Migration.from_name('V01_02__test2.sql', os.path.join('tests', 'data', 'schema'))]
    d = Utils.subtract(a, b)

    assert c[0].name == d[0].name


@pytest.mark.helpers_test
def test_subtract_noresult() -> None:
    a = [Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema'))]
    b = [Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema'))]
    c = Utils.subtract(a, b)

    assert c == []


@pytest.mark.helpers_test
def test_subtract_onlyonearray() -> None:
    a = [Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema'))]
    b = []
    c = Utils.subtract(a, b)

    assert c == a


@pytest.mark.helpers_test
def test_expected_pattern() -> None:
    pattern = Utils.expected_pattern()
    assert pattern == "V{major}_{minor}__{description}.sql"


@pytest.mark.helpers_test
def test_get_version_from_name() -> None:
    with pytest.raises(Exception):
        _ = Utils.get_version_from_name("test1.sql")
    assert True


@pytest.mark.helpers_test
def test_load_checksum_from_name_failed() -> None:
    with pytest.raises(Exception):
        _ = Utils.load_checksum_from_name('test', 'test')
    assert True


@pytest.mark.helpers_test
def test_version_name() -> None:
    assert Utils.is_file_name_valid('V1_1__test1.sql')


@pytest.mark.helpers_test
def test_semantic_version_name() -> None:
    assert Utils.is_file_name_valid('V1_0_1__test1.sql')


@pytest.mark.helpers_test
def test_semantic_version_name_major_period() -> None:
    assert Utils.is_file_name_valid('V1.0_1__test1.sql')


@pytest.mark.helpers_test
def test_semantic_version_name_minor_period() -> None:
    assert Utils.is_file_name_valid('V1_0.1__test1.sql')


@pytest.mark.helpers_test
def test_semantic_version_name_major_minor_period() -> None:
    assert Utils.is_file_name_valid('V1.0.1__test1.sql')


@pytest.mark.helpers_test
def test_semantic_version_name_minor_over_2digits() -> None:
    assert Utils.is_file_name_valid('V1_0_100__test1.sql')



import pytest
import os
from strip_ansi import strip_ansi
from pyway.info import Info
from pyway.settings import ConfigFile

from mysqld_integration_test import Mysqld

INFO_OUTPUT = """+-----------+-------------+-------------------+------------+-------------------+
|   version | extension   | name              | checksum   | apply_timestamp   |
|-----------+-------------+-------------------+------------+-------------------|
|      1.01 | SQL         | V01_01__test1.sql | new        | new               |
|      1.02 | SQL         | V01_02__test2.sql | new        | new               |
|      1.03 | SQL         | V01_03__test3.sql | new        | new               |
+-----------+-------------+-------------------+------------+-------------------+"""


@pytest.fixture
def mysqld_connect(autouse=True) -> Mysqld:
    mysqld = Mysqld()
    return mysqld.run()


@pytest.mark.info_test
@pytest.mark.mysqld_test
def test_pyway_info(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    tbl = Info(config).run()
    assert strip_ansi(tbl) == INFO_OUTPUT


@pytest.mark.info_test
@pytest.mark.mysqld_test
def test_pyway_info_nofiles(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    files = Info(config).get_new_local_migrations([], config.database_migration_dir)
    assert files == []



import pytest
import os
from strip_ansi import strip_ansi
from pyway.validate import Validate
from pyway.import_ import Import
from pyway.settings import ConfigFile

from pyway.dbms.database import factory

VALIDATE_OUTPUT = """Validating --> V01_01__test1.sql
V01_01__test1.sql VALID
"""


@pytest.fixture
def sqlite_connect(autouse: bool = True):
    # Delete any existing databases
    try:
        os.remove("./unittest-validate.sqlite")
    except Exception:
        pass

    args = ConfigFile()
    args.database_type = "sqlite"
    args.database_name = "./unittest-validate.sqlite"
    args.database_table = "pyway"

    return factory(args.database_type)(args)


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate(sqlite_connect) -> None:
    """ Import a file and validate that it matches """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    output = Import(config).run()
    output = Validate(config).run()
    assert strip_ansi(output) == VALIDATE_OUTPUT


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_noschemasfound(sqlite_connect) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    with pytest.raises(RuntimeError):
        _ = Validate(config).run()

    assert True


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_noschemasfound_skiperror(sqlite_connect) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    output = Validate(config).run(skip_initial_check=True)
    assert output == ""


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_nofilesfound(sqlite_connect) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    _ = Import(config).run()

    # Change to an empty directory for local files
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    with pytest.raises(RuntimeError):
        _ = Validate(config).run()

    assert True


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_diffname(sqlite_connect) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffname')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("with diff name of the database" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_diffchecksum(sqlite_connect) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffchecksum')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("with diff script" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_diffchecksum_dos(sqlite_connect) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffchecksum_dos')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("DOS" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.sqlite_test
def test_pyway_table_validate_outofdate(sqlite_connect) -> None:
    """ Import a file and remove that file """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-validate.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Import second file
    config.schema_file = "V01_02__test2.sql"
    _ = Import(config).run()

    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_outofdate-sqlite')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("Out of date" in str(e.value))



import os
from typing import Tuple

from pyway.helpers import Utils
from pyway.migration import Migration
from pyway.dbms.database import factory
from pyway.configfile import ConfigFile


class Checksum():
    def __init__(self, args: ConfigFile) -> None:
        self._db = factory(args.database_type)(args)
        self.migration_dir = args.database_migration_dir
        self.checksum_file = args.checksum_file
        self.args = args

    def run(self) -> Tuple[str, str]:
        if not self.checksum_file:
            raise AttributeError("Error, must specify --checksum-file with checksum")

        # If a path is specified, strip that off - all files should
        # be in the migration_dir directory
        if os.path.isabs(self.checksum_file) or os.sep in self.checksum_file:
            self.checksum_file = os.path.basename(self.checksum_file)

        if not os.path.exists(os.path.join(os.getcwd(), self.migration_dir, self.checksum_file)):
            raise FileNotFoundError(f"Error, schema file '{self.migration_dir}/{self.checksum_file}' does not exist!")

        # Generate new checksum
        version: str = Utils.get_version_from_name(self.checksum_file)
        migration: Migration = self._db.get_schema_migration(version)
        migration.checksum = Utils.load_checksum_from_name(self.checksum_file, self.migration_dir)

        self._db.update_checksum(migration)

        return self.checksum_file, migration.checksum



import pytest
import os
from strip_ansi import strip_ansi
from pyway.info import Info
from pyway.settings import ConfigFile

from pyway.dbms.database import factory

INFO_OUTPUT = """+-----------+-------------+-------------------+------------+-------------------+
|   version | extension   | name              | checksum   | apply_timestamp   |
|-----------+-------------+-------------------+------------+-------------------|
|      1.01 | SQL         | V01_01__test1.sql | new        | new               |
|      1.02 | SQL         | V01_02__test2.sql | new        | new               |
|      1.03 | SQL         | V01_03__test3.sql | new        | new               |
|      1.04 | SQL         | V01_04__test4.sql | new        | new               |
+-----------+-------------+-------------------+------------+-------------------+"""


@pytest.fixture
def sqlite_connect(autouse: bool = True):
    # Delete any existing databases
    try:
        os.remove("./unittest-info.sqlite")
    except Exception:
        pass

    args = ConfigFile()
    args.database_type = "sqlite"
    args.database_name = "./unittest-info.sqlite"
    args.database_table = "pyway"

    return factory(args.database_type)(args)


@pytest.mark.info_test
@pytest.mark.sqlite_test
def test_pyway_info(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-info.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    tbl = Info(config).run()
    assert strip_ansi(tbl) == INFO_OUTPUT


@pytest.mark.info_test
@pytest.mark.sqlite_test
def test_pyway_info_nofiles(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-info.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    files = Info(config).get_new_local_migrations([], config.database_migration_dir)
    assert files == []



import os
from typing import List, Any, Union

from pyway.helpers import bcolors
from pyway.helpers import Utils
from pyway.dbms.database import factory
from pyway.migration import Migration
from pyway.errors import (OUT_OF_DATE_ERROR, DIFF_NAME_ERROR, DIFF_CHECKSUM_ERROR,
                          MIGRATIONS_NOT_FOUND, MIGRATIONS_NOT_STARTED,
                          DIFF_CHECKSUM_ERROR_DOS)
from pyway.configfile import ConfigFile


class Validate():
    def __init__(self, args: ConfigFile) -> None:
        self._db = factory(args.database_type)(args)
        self.migration_dir = args.database_migration_dir
        self.args = args

    def run(self,  skip_initial_check: bool = False) -> str:
        local_migrations = self._get_all_local_migrations()
        db_migrations = self._db.get_all_schema_migrations()
        output = ""

        if not db_migrations:
            if not skip_initial_check:
                raise RuntimeError(MIGRATIONS_NOT_STARTED)

        if db_migrations and not local_migrations:
            if not skip_initial_check:
                raise RuntimeError(MIGRATIONS_NOT_FOUND % self.migration_dir)

        if local_migrations:
            local_migrations_map = Utils.create_map_from_list("version", local_migrations)
            for db_migration in db_migrations:
                output += Utils.color(f"Validating --> {db_migration.name}\n", bcolors.OKBLUE)
                local_migration: Union[Migration, Any] = local_migrations_map.get(db_migration.version)
                if self._out_of_date(local_migration):
                    raise RuntimeError(OUT_OF_DATE_ERROR % db_migration.name)
                elif not self._diff_names(local_migration, db_migration):
                    raise RuntimeError(DIFF_NAME_ERROR % (local_migration.name, db_migration.name))
                elif not self._diff_checksum(local_migration, db_migration):
                    if self._has_dos_line_endings(os.path.join(os.getcwd(), self.migration_dir, local_migration.name)):
                        raise RuntimeError(DIFF_CHECKSUM_ERROR_DOS % (local_migration.name,
                                                                      local_migration.checksum,
                                                                      db_migration.checksum))
                    else:
                        raise RuntimeError(DIFF_CHECKSUM_ERROR % (local_migration.name,
                                                                  local_migration.checksum,
                                                                  db_migration.checksum))
                else:
                    output += Utils.color(f"{db_migration.name} VALID\n", bcolors.OKGREEN)

        return output

    def _out_of_date(self, local_migration: Migration) -> bool:
        return bool(local_migration is None)

    def _diff_names(self, local_migration: Migration, db_migration: Migration) -> bool:
        return bool(local_migration.name == db_migration.name)

    def _diff_checksum(self, local_migration: Migration, db_migration: Migration) -> bool:
        return bool(local_migration.checksum == db_migration.checksum)

    def _get_all_local_migrations(self) -> List:
        local_files = Utils.get_local_files(self.migration_dir)
        if not local_files:
            return []
        migrations = [Migration.from_name(local_file, self.migration_dir) for local_file in local_files]
        return Utils.sort_migrations_list(migrations)

    def _has_dos_line_endings(self, file_path: str) -> bool:
        with open(file_path, 'rb') as file:
            for line in file:
                if b'\r\n' in line:
                    return True
        return False



import sqlite3
from typing import Any, List, Tuple

from pyway.migration import Migration
from pyway.configfile import ConfigFile


CREATE_VERSION_MIGRATIONS = "create table if not exists %s ("\
    "installed_rank serial PRIMARY KEY,"\
    "version varchar(20) NOT NULL,"\
    "extension varchar(20) NOT NULL,"\
    "name varchar(125) NOT NULL,"\
    "checksum varchar(25) NOT NULL,"\
    "apply_timestamp timestamp DEFAULT CURRENT_TIMESTAMP"\
    ");"
SELECT_FIELDS = ("version", "extension", "name", "checksum", "apply_timestamp")
ORDER_BY_FIELD_ASC = "installed_rank"
ORDER_BY_FIELD_DESC = "installed_rank desc"
INSERT_VERSION_MIGRATE = "insert into %s (version, extension, name, checksum) values ('%s', '%s', '%s', '%s');"
UPDATE_CHECKSUM = "update %s set checksum='%s' where version='%s';"


class Sqlite():

    def __init__(self, config: ConfigFile) -> None:
        self.config = config
        self.version_table = config.database_table
        self.create_version_table_if_not_exists()

    def connect(self) -> Any:
        conn = sqlite3.connect(self.config.database_name)
        return conn

    def create_version_table_if_not_exists(self) -> None:
        self.execute(CREATE_VERSION_MIGRATIONS % self.version_table)

    def execute(self, script: str) -> List[Tuple]:
        cnx = self.connect()
        cursor = cnx.cursor()
        cursor.executescript(script)
        rows = cursor.fetchall()
        cnx.commit()
        cnx.close()
        return rows

    def get_all_schema_migrations(self) -> List[Migration]:
        cnx = self.connect()
        cursor = cnx.cursor()
        cursor.execute(f"SELECT {','.join(SELECT_FIELDS)} FROM {self.version_table} ORDER BY {ORDER_BY_FIELD_ASC}")
        migrations = []
        for row in cursor.fetchall():
            migrations.append(Migration(row[0], row[1], row[2], row[3], row[4]))
        cursor.close()
        cnx.close()
        return migrations

    def get_schema_migration(self, version: str) -> Migration:
        cnx = self.connect()
        cursor = cnx.cursor()
        cursor.execute(f"SELECT {','.join(SELECT_FIELDS)} FROM {self.version_table} WHERE version='{version}'")
        row = cursor.fetchone()
        if row is not None:
            migration = Migration(row[0], row[1], row[2], row[3], row[4])
        cursor.close()
        cnx.close()
        return migration

    def upgrade_version(self, migration: Migration) -> None:
        self.execute(INSERT_VERSION_MIGRATE % (self.version_table, migration.version,
                                               migration.extension, migration.name,
                                               migration.checksum))

    def update_checksum(self, migration: Migration) -> None:
        self.execute(UPDATE_CHECKSUM % (self.version_table, migration.checksum, migration.version))



import os
from typing import Any, Union


class ConfigFile():
    def __init__(self, **kwargs: Any) -> None:
        self.database_migration_dir = os.environ.get('PYWAY_DATABASE_MIGRATION_DIR', 'resources')
        self.database_table = os.environ.get('PYWAY_TABLE', kwargs.get('database_table'))
        self.database_type = os.environ.get('PYWAY_TYPE', kwargs.get('database_type'))
        self.database_host = os.environ.get('PYWAY_DATABASE_HOST', kwargs.get('database_host'))
        self.database_port = os.environ.get('PYWAY_DATABASE_PORT', kwargs.get('database_port'))
        self.database_name = os.environ.get('PYWAY_DATABASE_NAME', kwargs.get('database_name'))
        self.database_username = os.environ.get('PYWAY_DATABASE_USERNAME', kwargs.get('database_username'))
        self.database_password = os.environ.get('PYWAY_DATABASE_PASSWORD', kwargs.get('database_password'))
        self.schema_file: Union[str, None] = None
        self.checksum_file = None
        self.config = os.environ.get('PYWAY_CONFIG_FILE', '.pyway.conf')
        self.version = False
        self.cmd = None


class MockConfig():
    pass


class MockArgs():
    pass



import pytest
import os
import sys
from pyway.settings import ConfigFile
from pyway.settings import Settings
from pyway.settings import ARGS
from pyway.configfile import MockConfig
from pyway.configfile import MockArgs


# Make sure config options exists and check some defaults
@pytest.mark.settings_test
def test_settings_database_migration_dir() -> None:
    config = ConfigFile()
    assert config.database_migration_dir == 'resources'


@pytest.mark.settings_test
def test_settings_database_table() -> None:
    config = ConfigFile(database_table='public.pyway')
    assert config.database_table == 'public.pyway'


@pytest.mark.settings_test
def test_settings_database_type() -> None:
    config = ConfigFile(database_type='postgres')
    assert config.database_type == 'postgres'


@pytest.mark.settings_test
def test_settings_database_host() -> None:
    config = ConfigFile(database_host='localhost')
    assert config.database_host == 'localhost'


@pytest.mark.settings_test
def test_settings_database_port() -> None:
    config = ConfigFile(database_port='5432')
    assert config.database_port == '5432'


@pytest.mark.settings_test
def test_settings_database_name() -> None:
    config = ConfigFile(database_name='postgres')
    assert config.database_name == 'postgres'


@pytest.mark.settings_test
def test_settings_database_username() -> None:
    config = ConfigFile(database_username='postgres')
    assert config.database_username == 'postgres'


@pytest.mark.settings_test
def test_settings_database_password() -> None:
    config = ConfigFile(database_password='password')
    assert config.database_password == 'password'


@pytest.mark.settings_test
def test_settings_schema_file() -> None:
    config = ConfigFile()
    assert config.schema_file is None


@pytest.mark.settings_test
def test_settings_config() -> None:
    config = ConfigFile()
    assert config.config == '.pyway.conf'


@pytest.mark.settings_test
def test_settings_version() -> None:
    config = ConfigFile()
    assert config.version is False


@pytest.mark.settings_test
def test_settings_cmd() -> None:
    config = ConfigFile()
    assert config.cmd is None


@pytest.mark.settings_test
def test_parse_config_file() -> None:
    config = ConfigFile()
    config.config = 'tests/data/pyway.conf'
    config = Settings.parse_config_file(config)
    assert config.database_username == "unittest"


@pytest.mark.settings_test
def test_parse_args() -> None:

    # Setup mock config and args
    config = MockConfig()
    args = MockArgs()

    # Set attributes on args for testing
    for arg in ARGS:
        setattr(args, arg, f"test_{arg}")

    # Call your function
    Settings.parse_args(config, args)

    # Assert that config is updated correctly
    for arg in ARGS:
        assert getattr(config, arg) == f"test_{arg}"


@pytest.mark.settings_test
def test_parse_arguments() -> None:
    class MockConfig:
        pass

    # Create a mock config object
    config = MockConfig()

    test_args = [
        'script_name',
        '--database-migration-dir', 'migrations',
        '--database-table', 'pyway_meta',
        '--database-type', 'postgres',
        '--database-host', 'localhost'
    ]

    _ = sys.argv
    sys.argv = test_args

    Settings.parse_arguments(config)

    assert config.database_migration_dir == 'migrations'
    assert config.database_table == 'pyway_meta'
    assert config.database_type == 'postgres'
    assert config.database_host == 'localhost'


@pytest.mark.settings_test
def test_env_var_interpolation() -> None:
    # Set an env var
    os.environ['TEST_VAR'] = 'sometest'

    config = ConfigFile()
    config.config = 'tests/data/pyway_variable.conf'
    config = Settings.parse_config_file(config)
    assert config.database_username == "unittest_sometest"



import pytest
import os
from pyway.checksum import Checksum
from pyway.migrate import Migrate
from pyway.settings import ConfigFile

from postgresql_integration_test import PostgreSQL


@pytest.fixture
def postgresql_connect(autouse: bool = True) -> PostgreSQL:
    postgresql = PostgreSQL()
    return postgresql.run()


@pytest.mark.checksum_test
@pytest.mark.postgresql_test
def test_pyway_table_checksum(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-postgres')
    config.checksum_file = "V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "B78E2BE3"


@pytest.mark.checksum_test
@pytest.mark.postgresql_test
def test_pyway_table_checksum_fileinvalid(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-postgres')

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(AttributeError):
        _, _ = Checksum(config).run()

    assert True


@pytest.mark.checksum_test
@pytest.mark.postgresql_test
def test_pyway_table_checksum_fullpath(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-postgres')
    config.checksum_file = "schema/V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "B78E2BE3"


@pytest.mark.checksum_test
@pytest.mark.postgresql_test
def test_pyway_table_checksum_invalid_filename(postgresql_connect: PostgreSQL) -> None:
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'public.pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-postgres')
    config.checksum_file = "invalidfilename.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(FileNotFoundError):
        _, _ = Checksum(config).run()

    assert True



import pytest
import os
from pyway.checksum import Checksum
from pyway.migrate import Migrate
from pyway.settings import ConfigFile

from pyway.dbms.database import factory


@pytest.fixture
def sqlite_connect(autouse: bool = True):
    # Delete any existing databases
    try:
        os.remove("./unittest-checksum.sqlite")
    except Exception:
        pass

    args = ConfigFile()
    args.database_type = "sqlite"
    args.database_name = "./unittest-checksum.sqlite"
    args.database_table = "pyway"

    return factory(args.database_type)(args)


@pytest.mark.checksum_test
@pytest.mark.sqlite_test
def test_pyway_table_checksum(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = "./unittest-checksum.sqlite"
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.checksum_file = "V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "9B8E1E62"


@pytest.mark.checksum_test
@pytest.mark.sqlite_test
def test_pyway_table_checksum_fileinvalid(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-checksum.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(AttributeError):
        _, _ = Checksum(config).run()

    assert True


@pytest.mark.checksum_test
@pytest.mark.sqlite_test
def test_pyway_table_checksum_fullpath(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-checksum.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.checksum_file = "schema/V01_01__test1.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    name, checksum = Checksum(config).run()
    assert name == "V01_01__test1.sql"
    assert checksum == "9B8E1E62"


@pytest.mark.checksum_test
@pytest.mark.sqlite_test
def test_pyway_table_checksum_invalid_filename(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-checksum.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.checksum_file = "invalidfilename.sql"

    # Add migration
    _ = Migrate(config).run()

    # Test once migration is complete
    with pytest.raises(FileNotFoundError):
        _, _ = Checksum(config).run()

    assert True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import annotations
import pytest

from pyway.configfile import ConfigFile
from pyway.dbms.database import factory
from pyway.dbms import duckdb
from pyway.migration import Migration


@pytest.mark.duckdb_test
def test_factory() -> None:
    args = ConfigFile()
    args.database_type = "duckdb"
    args.database_name = "./unittest.duckdb"
    args.database_table = "pyway"
    db: duckdb.Duckdb = factory(args.database_type)(args)

    assert db.connect().sql(
        f"select 1 from information_schema.tables where table_name = '{args.database_table}'"
    ).fetchone() is not None

    db.disconnect()


@pytest.mark.duckdb_test
def test_migrations() -> None:
    args = ConfigFile()
    args.database_type = "duckdb"
    args.database_name = "./unittest.duckdb"
    args.database_table = "pyway"
    db: duckdb.Duckdb = factory(args.database_type)(args)

    db.execute(f"truncate table {args.database_table}")

    migrations = db.get_all_schema_migrations()
    assert migrations == []

    mig = Migration(
        version="01.01", extension="SQL",
        name="V01_01__test1.sql",
        checksum="8327AD7B",
        apply_timestamp="new"
    )

    db.upgrade_version(mig)

    migrations = db.get_all_schema_migrations()
    assert len(migrations) == 1
    assert migrations[0].checksum == mig.checksum

    updated = Migration(
        version="01.01", extension="SQL",
        name="V01_01__test1.sql",
        checksum="FACB0AE4",
        apply_timestamp="new"
    )

    db.update_checksum(updated)
    migrations = db.get_all_schema_migrations()
    assert len(migrations) == 1
    assert migrations[0].checksum == updated.checksum

    fetched = db.get_schema_migration(updated.version)
    assert fetched.version == updated.version
    assert fetched.name == updated.name
    assert fetched.checksum == updated.checksum
    assert fetched.extension == updated.extension

    db.disconnect()



import pytest
import os
from pyway.migration import Migration


@pytest.mark.migration_test
def test_from_list() -> None:
    migrations = []
    migration = Migration('01.01', 'SQL', 'V01_01__testfile.sql', '53175082', '2023-02-28 15:56:14')
    migrations.append(migration)
    test_list = Migration.from_list(migrations)
    assert test_list[0].name == migration.name


@pytest.mark.migration_test
def test_from_name() -> None:
    migration = Migration.from_name('V01_01__test1.sql', os.path.join('tests', 'data', 'schema'))
    assert migration.name == 'V01_01__test1.sql'
    assert migration.version == "01.01"


@pytest.mark.migration_test
def test_semantic_version_from_name() -> None:
    migration = Migration.from_name('V01_01_01__test1.sql', os.path.join('tests', 'data', 'schemasemver'))
    assert migration.name == 'V01_01_01__test1.sql'
    assert migration.version == "01.01.01"


@pytest.mark.migration_test
def test_str() -> None:
    migration = Migration.from_name('V01_01_01__test1.sql', os.path.join('tests', 'data', 'schemasemver'))
    assert str(migration) == "version=01.01.01, extension=SQL, name=V01_01_01__test1.sql, " \
                             "checksum=8327AD7B, apply_timestamp=None"



import os
import argparse
import yaml
from typing import Tuple, Dict, Union

from pyway.configfile import ConfigFile
from pyway.configfile import MockArgs

# Pyway consts
SQL_MIGRATION_PREFIX = os.environ.get('PYWAY_SQL_MIGRATION_PREFIX', 'V')
SQL_MIGRATION_SEPARATOR = os.environ.get('PYWAY_SQL_MIGRATION_SEPARATOR', '__')
SQL_MIGRATION_SUFFIXES = os.environ.get('PYWAY_SQL_MIGRATION_SUFFIXES', '.sql')
ARGS = ['database_migration_dir', 'database_table', 'database_type', 'database_host',
        'database_port', 'database_name', 'database_username', 'database_password',
        'schema_file', 'checksum_file', 'config', 'version', 'cmd']


class Settings():
    @staticmethod
    def parse_args(config: ConfigFile, args: Union[argparse.Namespace, MockArgs]) -> ConfigFile:
        for arg in ARGS:
            if getattr(args, arg):
                setattr(config, arg, getattr(args, arg))
        return config

    @classmethod
    def parse_arguments(self, config: ConfigFile) -> Tuple[ConfigFile, argparse.ArgumentParser]:
        parser: argparse.ArgumentParser = argparse.ArgumentParser()
        parser.add_argument("--database-migration-dir", help="Database migration directory")
        parser.add_argument("--database-table", help="Database table that stores pyway metadata")
        parser.add_argument("--database-type", help="Database type [postgres|mysql|duckdb|sqlite]")
        parser.add_argument("--database-host", help="Database host")
        parser.add_argument("--database-port", help="Database port")
        parser.add_argument("--database-name", help="Database name")
        parser.add_argument("--database-username", help="Database username")
        parser.add_argument("--database-password", help="Database password")

        parser.add_argument("--schema-file", help="Schema file for import")
        parser.add_argument("--checksum-file", help="Checksum to update")
        parser.add_argument("-c", "--config", help="Config file")
        parser.add_argument("-v", "--version", help="Version", action='store_true')
        parser.add_argument("cmd", nargs="?", help="info|validate|migrate|import|checksum")

        new_config: ConfigFile = self.parse_args(config, parser.parse_args())
        return (new_config, parser)

    @classmethod
    def parse_config_file(self, config: ConfigFile) -> ConfigFile:
        # See if there is a config file
        if os.path.exists(config.config):
            with open(config.config, "r", encoding='utf-8') as ymlfile:
                cfg: Dict = yaml.load(ymlfile, Loader=yaml.FullLoader)

            # Merge config together with args
            for c in cfg:
                if isinstance(cfg[c], str):
                    # Interpolate env vars
                    cfg[c] = os.path.expandvars(cfg[c])
                setattr(config, c, cfg[c])

        return config



import sys
import logging
from pyway.helpers import Utils
from pyway.helpers import bcolors
from pyway.exceptions import InvalidLogLevel


class _Log():
    def __init__(self) -> None:
        self.logger = logging.getLogger('pyway')
        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        self.logger.setLevel(logging.INFO)

    def debug(self, msg: str) -> None:
        if self.logger:
            self.logger.debug(msg)

    def info(self, msg: str) -> None:
        if self.logger:
            self.logger.info(Utils.color(msg, bcolors.OKBLUE))

    def error(self, msg: str) -> None:
        if self.logger:
            self.logger.error(Utils.color(msg, bcolors.FAIL))

    def success(self, msg: str) -> None:
        self.logger.info(Utils.color(msg, bcolors.OKGREEN))

    def setlevel(self, log_level: str) -> None:
        if log_level == "INFO":
            self.logger.setLevel(logging.INFO)
        elif log_level == "DEBUG":
            self.logger.setLevel(logging.DEBUG)
        elif log_level == 'ERROR':
            self.logger.setLevel(logging.ERROR)
        elif log_level == 'WARN':
            self.logger.setLevel(logging.WARN)
        else:
            raise InvalidLogLevel


logger = _Log()



import pytest
import os
from strip_ansi import strip_ansi
from pyway.validate import Validate
from pyway.import_ import Import
from pyway.settings import ConfigFile

from postgresql_integration_test import PostgreSQL

VALIDATE_OUTPUT = """Validating --> V01_01__test1.sql
V01_01__test1.sql VALID
"""


@pytest.fixture
def postgresql_connect(autouse: bool = True) -> PostgreSQL:
    postgresql = PostgreSQL()
    return postgresql.run()


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate(postgresql_connect: PostgreSQL) -> None:
    """ Import a file and validate that it matches """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    output = Import(config).run()
    output = Validate(config).run()
    assert strip_ansi(output) == VALIDATE_OUTPUT


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_noschemasfound(postgresql_connect: PostgreSQL) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    with pytest.raises(RuntimeError):
        _ = Validate(config).run()

    assert True


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_noschemasfound_skiperror(postgresql_connect: PostgreSQL) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    output = Validate(config).run(skip_initial_check=True)
    assert output == ""


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_nofilesfound(postgresql_connect: PostgreSQL) -> None:
    """ Test to see what happens when we try to validate and no files are found """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    _ = Import(config).run()

    # Change to an empty directory for local files
    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    with pytest.raises(RuntimeError):
        _ = Validate(config).run()

    assert True


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_diffname(postgresql_connect: PostgreSQL) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffname')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("with diff name of the database" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_diffchecksum(postgresql_connect: PostgreSQL) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-postgres')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffchecksum')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("with diff script" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_diffchecksum_dos(postgresql_connect: PostgreSQL) -> None:
    """ Import a file and change the filename """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Change the filename
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_diffchecksum_dos')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("DOS" in str(e.value))


@pytest.mark.validate_test
@pytest.mark.postgresql_test
def test_pyway_table_validate_outofdate(postgresql_connect: PostgreSQL) -> None:
    """ Import a file and remove that file """
    config = ConfigFile()
    config.database_type = "postgres"
    config.database_host = postgresql_connect.host
    config.database_username = postgresql_connect.username
    config.database_port = postgresql_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    # Import file
    _ = Import(config).run()

    # Import second file
    config.schema_file = "V01_02__test2.sql"
    _ = Import(config).run()

    # Change to empty dir
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_validate_outofdate')

    with pytest.raises(RuntimeError) as e:
        _ = Validate(config).run()

    assert bool("Out of date" in str(e.value))



import pytest
import os
from pyway.import_ import Import
from pyway.settings import ConfigFile

from pyway.dbms.database import factory

INFO_OUTPUT = """+-----------+-------------+-------------------+------------+-------------------+
|   version | extension   | name              | checksum   | apply_timestamp   |
|-----------+-------------+-------------------+------------+-------------------|
|      1.01 | SQL         | V01_01__test1.sql | new        | new               |
|      1.02 | SQL         | V01_02__test2.sql | new        | new               |
|      1.03 | SQL         | V01_03__test3.sql | new        | new               |
+-----------+-------------+-------------------+------------+-------------------+"""

VALIDATE_OUTPUT = """Validating --> V01_01__test1.sql
V01_01__test1.sql VALID
"""


@pytest.fixture
def sqlite_connect(autouse: bool = True):
    # Delete any existing databases
    try:
        os.remove("./unittest-import.sqlite")
    except Exception:
        pass

    args = ConfigFile()
    args.database_type = "sqlite"
    args.database_name = "./unittest-import.sqlite"
    args.database_table = "pyway"

    return factory(args.database_type)(args)


@pytest.mark.import_test
@pytest.mark.sqlite_test
def test_pyway_table_import(sqlite_connect) -> None:
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-import.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = "V01_01__test1.sql"
    output = Import(config).run()
    assert output == "V01_01__test1.sql"


@pytest.mark.import_test
@pytest.mark.sqlite_test
def test_pyway_table_import_fullfilepath(sqlite_connect) -> None:
    """ Schema file is specified with path """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-import.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = f"{config.database_migration_dir}/V01_01__test1.sql"
    output = Import(config).run()
    assert output == "V01_01__test1.sql"


@pytest.mark.import_test
@pytest.mark.sqlite_test
def test_pyway_table_import_noschema(sqlite_connect) -> None:
    """ schema_file is missing from arguments """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-import.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    with pytest.raises(AttributeError):
        _ = Import(config).run()
    assert True


@pytest.mark.import_test
@pytest.mark.sqlite_test
def test_pyway_table_import_filenotfound(sqlite_connect) -> None:
    """ Schema file specified does not exist in migration_dir """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-import.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema-sqlite')
    config.schema_file = "V01_01__test1notfound.sql"
    with pytest.raises(FileNotFoundError):
        _ = Import(config).run()
    assert True


@pytest.mark.import_test
@pytest.mark.sqlite_test
def test_pyway_table_import_invalidfilename(sqlite_connect) -> None:
    """ Schema file exists but is not named properly """
    config = ConfigFile()
    config.database_type = "sqlite"
    config.database_name = './unittest-import.sqlite'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema_invalid_file')
    config.schema_file = "invalidfilename.sql"
    with pytest.raises(ValueError):
        _ = Import(config).run()
    assert True



import os
import re
import zlib
from typing import Any, Dict, List, Iterable

from pyway import settings
from pyway.errors import VALID_NAME_ERROR, DIRECTORY_NOT_FOUND, OUT_OF_DATE_ERROR


class bcolors():
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Utils():

    @staticmethod
    def subtract(list_a: List, list_b: List) -> List:
        result = []
        if list_a and list_b:
            checksum_list_b = [b.checksum for b in list_b]
            result = [a for a in list_a if a.checksum not in checksum_list_b]
        elif list_a and not list_b:
            # List B is empty (usually from a new install)
            return list_a
        return result

    @staticmethod
    def expected_pattern() -> str:
        return f'{settings.SQL_MIGRATION_PREFIX}{{major}}_{{minor}}{settings.SQL_MIGRATION_SEPARATOR}' \
                f'{{description}}{settings.SQL_MIGRATION_SUFFIXES}'

    @staticmethod
    def is_file_name_valid(name: str) -> bool:
        _pattern = r"%s\d+[._]\d+|\d+[._]\d+__%s\.%s$" % \
            (settings.SQL_MIGRATION_PREFIX, settings.SQL_MIGRATION_SEPARATOR, settings.SQL_MIGRATION_SUFFIXES)
        return re.match(_pattern, name, re.IGNORECASE) is not None

    @staticmethod
    def sort_migrations_list(migrations: List[Any]) -> List[Any]:
        return sorted(migrations, key=lambda x: [x.get("version"), x.get("name")] if isinstance(x, dict) else
                                                [x.version, x.name], reverse=False)

    @staticmethod
    def flatten_migrations(migrations: Iterable[Any]) -> List[Dict[Any, Any]]:
        migration_list = []
        for migration in migrations:
            migration_list.append({'version': migration.version, 'extension': migration.extension,
                                   'name': migration.name, 'checksum': migration.checksum,
                                   'apply_timestamp': migration.apply_timestamp})
        return migration_list

    @staticmethod
    def get_version_from_name(name: str) -> str:
        ver = re.findall(r"(\d+[._]\d+|\d+)[._](\d+)__", name)
        if not ver:
            raise ValueError(VALID_NAME_ERROR % (name, Utils.expected_pattern()))

        if isinstance(ver[0], tuple):
            version: str = ".".join(ver[0])
            version = version.replace("_", ".")
        else:
            version = ver[0].replace("_", ".")

        return version

    @staticmethod
    def get_extension_from_name(name: str) -> str:
        return name.split('.')[-1].upper()

    @staticmethod
    def load_checksum_from_name(name: str, path: str) -> str:
        fullname = os.path.join(os.getcwd(), path, name)
        prev = 0
        try:
            for line in open(fullname, "rb"):
                prev = zlib.crc32(line, prev)
            return "%X" % (prev & 0xFFFFFFFF)
        except FileNotFoundError:
            raise FileNotFoundError(OUT_OF_DATE_ERROR % fullname.split("/")[-1])

    @staticmethod
    def basepath(d: str) -> str:
        return os.path.join(os.getcwd(), d)

    @staticmethod
    def get_local_files(d: str) -> List[str]:
        path = Utils.basepath(d)
        dir_list = []
        try:
            # Skip any hidden files
            for f in os.listdir(path):
                if not f.startswith('.'):
                    dir_list.append(f)
        except OSError:
            raise FileNotFoundError(DIRECTORY_NOT_FOUND % path)
        return dir_list

    @staticmethod
    def create_map_from_list(key: str, list_: List[Any]) -> Dict[Any, Any]:
        return {lst.__dict__[key]: lst for lst in list_}

    @staticmethod
    def color(msg: str, color: str) -> str:
        return f"{color}{msg}{bcolors.ENDC}"



import pytest
from pyway.version import __version__


@pytest.mark.version_test
def test_version_isnotnone() -> None:
    assert __version__ is not None


@pytest.mark.version_test
def test_version_semver() -> None:
    assert len(__version__.split('.')) == 3



import pytest
import os
from strip_ansi import strip_ansi
from pyway.migrate import Migrate
from pyway.settings import ConfigFile

from mysqld_integration_test import Mysqld

MIGRATE_OUTPUT = """Migrating --> V01_01__test1.sql
V01_01__test1.sql SUCCESS
Migrating --> V01_02__test2.sql
V01_02__test2.sql SUCCESS
Migrating --> V01_03__test3.sql
V01_03__test3.sql SUCCESS
"""


MIGRATE_OUTPUT_NOTHING = """Nothing to do
"""


@pytest.fixture
def mysqld_connect(autouse: bool = True) -> Mysqld:
    mysqld = Mysqld()
    return mysqld.run()


@pytest.mark.migrate_test
@pytest.mark.mysqld_test
def test_pyway_migrate(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')

    output = Migrate(config).run()
    assert strip_ansi(output) == MIGRATE_OUTPUT


@pytest.mark.migrate_test
@pytest.mark.mysqld_test
def test_pyway_migrate_nothingtodo(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')

    # Double migration to validate nothing
    output = Migrate(config).run()
    output = Migrate(config).run()

    assert strip_ansi(output) == MIGRATE_OUTPUT_NOTHING


@pytest.mark.migrate_test
@pytest.mark.mysqld_test
def test_pyway_migrate_no_local_files(mysqld_connect: Mysqld) -> None:
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data', 'schema')
    config.schema_file = "V01_01__test1.sql"

    _ = Migrate(config).run()

    config.database_migration_dir = os.path.join('tests', 'data', 'empty')

    # Double migration to validate nothing
    with pytest.raises(RuntimeError) as e:
        _ = Migrate(config).run()

    assert bool("no local migration files found" in str(e.value))
